from json import dumps as _dumps
from typing import Any as _Any, override as _override

from leads.data import Serializable


class ConfigTemplate(Serializable):
    def __init__(self, base: dict[str, _Any]) -> None:
        """
        All custom columns should be public (not named after "_").
        :param base: base dictionary
        """
        self._d: dict[str, _Any] = base
        self.debug_level: str = "DEBUG"
        self.srw_mode: bool = True
        self.refresh()

    def __getitem__(self, name: str) -> _Any | None:
        return self.get(name)

    def __getattr__(self, name: str) -> _Any | None:
        return self.get(name)

    def __setitem__(self, name: str, value: _Any) -> None:
        self._d[name] = value
        self.refresh()

    @_override
    def __str__(self) -> str:
        return _dumps(self.to_dict())

    def load(self, d: dict[str, _Any]) -> None:
        """
        Load a new dictionary and refresh.
        :param d: the dictionary to load
        """
        self._d = d
        self.refresh()

    def get(self, name: str, default: _Any | None = None) -> _Any:
        """
        Get the value of a given name from the dictionary.
        :param name: dictionary key
        :param default: default value if the value does not exist
        :return: the value if it exists or else the default value
        """
        if r := self._d.get(name):
            return r
        return default

    def refresh(self) -> None:
        """
        Write the dictionary into the instance attributes.
        """
        for name in dir(self):
            if not name.startswith("_") and (v := self.get(name)) is not None:
                setattr(self, name, v)



from typing import override as _override

from serial import Serial as _Serial

from leads.comm import ConnectionBase as _Connection, Service as _Service


class SerialConnection(_Connection):
    def __init__(self,
                 service: _Service,
                 serial: _Serial,
                 port: str,
                 remainder: bytes = b"",
                 separator: bytes = b";") -> None:
        super().__init__(service, remainder, separator)
        self._serial: _Serial = serial
        self._port: str = port

    @_override
    def closed(self) -> bool:
        return self._serial.closed

    def _require_open_serial(self, mandatory: bool = True) -> _Serial:
        if mandatory and self.closed():
            raise IOError("An open serial is required")
        return self._serial

    @_override
    def receive(self, chunk_size: int = 1) -> bytes | None:
        if self._remainder != b"":
            return self.use_remainder()
        try:
            msg = chunk = b""
            while self._separator not in chunk:
                msg += (chunk := self._require_open_serial().read(chunk_size))
            return self.with_remainder(msg)
        except IOError:
            return

    @_override
    def send(self, msg: bytes) -> None:
        self._require_open_serial().write(msg + self._separator)
        if msg == b"disconnect":
            self.close()

    @_override
    def close(self) -> None:
        self._require_open_serial(False).close()



from threading import Lock as _Lock
from typing import override as _override

from leads import Device as _Device


class Odometer(_Device):
    def __init__(self):
        super().__init__()
        self._milage: float = 0

    @_override
    def write(self, payload: float) -> None:
        self._milage += payload

    @_override
    def read(self) -> float:
        return self._milage


class ConcurrentOdometer(Odometer):
    def __init__(self):
        super().__init__()
        self._lock: _Lock = _Lock()

    @_override
    def write(self, payload: float) -> None:
        self._lock.acquire()
        try:
            super().write(payload)
        finally:
            self._lock.release()



from math import lcm as _lcm
from typing import Callable as _Callable, Self as _Self, TypeVar as _TypeVar, Generic as _Generic

from PIL import ImageTk as _ImageTk
from customtkinter import CTk as _CTk, CTkBaseClass as _CTkBaseClass

from leads_gui.runtime import RuntimeData
from leads_gui.system import _ASSETS_PATH, get_system_platform

type Widget = _CTkBaseClass

T = _TypeVar("T", bound=RuntimeData)


class Window(_Generic[T]):
    def __init__(self,
                 width: int,
                 height: int,
                 refresh_rate: int,
                 runtime_data: T,
                 on_refresh: _Callable[[_Self], None] = lambda _: None,
                 on_kill: _Callable[[_Self], None] = lambda _: None,
                 title: str = "LEADS",
                 fullscreen: bool = True,
                 no_title_bar: bool = True) -> None:
        self._root: _CTk = _CTk()
        self._root.title(title)
        self._root.wm_iconbitmap()
        self._root.iconphoto(True, _ImageTk.PhotoImage(file=_ASSETS_PATH + "/logo.png"))
        self._root.overrideredirect(no_title_bar)
        self._width: int = self._root.winfo_screenwidth() if fullscreen else width
        self._height: int = self._root.winfo_screenheight() if fullscreen else height
        self._root.geometry(str(self._width) + "x" + str(self._height))
        self._refresh_rate: int = refresh_rate
        self._refresh_interval: int = int(1000 / refresh_rate)
        self._runtime_data: T = runtime_data
        self._on_refresh: _Callable[[_Self], None] = on_refresh
        self._on_kill: _Callable[[_Self], None] = on_kill

        self._active: bool = False

    def root(self) -> _CTk:
        return self._root

    def width(self) -> int:
        return self._width

    def height(self) -> int:
        return self._height

    def refresh_rate(self) -> int:
        return self._refresh_rate

    def runtime_data(self) -> T:
        return self._runtime_data

    def set_on_refresh(self, on_refresh: _Callable[[_Self], None]) -> None:
        self._on_refresh = on_refresh

    def set_on_close(self, on_close: _Callable[[_Self], None]) -> None:
        self._on_kill = on_close

    def active(self) -> bool:
        return self._active

    def show(self) -> None:
        self._active = True

        def wrapper() -> None:
            self._on_refresh(self)
            if self._active:
                self._root.after(self._refresh_interval, wrapper)

        self._root.after(0, wrapper)
        self._root.mainloop()

    def kill(self) -> None:
        self._active = False
        self._root.destroy()


class ContextManager(object):
    def __init__(self, window: Window) -> None:
        self._window: Window = window
        self._widgets: dict[str, Widget] = {}
        self._system_platform: str = get_system_platform()

    def __setitem__(self, key: str, widget: Widget) -> None:
        self._widgets[key] = widget

    def __getitem__(self, key: str) -> Widget:
        return self._widgets[key]

    def system_platform(self) -> str:
        return self._system_platform

    def set(self, key: str, widget: Widget) -> None:
        self[key] = widget

    def get(self, key: str) -> Widget:
        return self[key]

    def parse_layout(self, layout: list[list[str | Widget]]) -> list[list[Widget]]:
        for i in range(len(layout)):
            for j in range(len(layout[i])):
                e = layout[i][j]
                if isinstance(e, str):
                    layout[i][j] = self[e]
        return layout

    def layout(self, layout: list[list[str | Widget]]) -> None:
        layout = self.parse_layout(layout)
        self.root().grid_columnconfigure(tuple(range(t := _lcm(*tuple(map(len, layout))))), weight=1)
        for i in range(len(layout)):
            row = layout[i]
            length = len(row)
            for j in range(length):
                s = int(t / length)
                row[j].grid(row=i, column=j * s, sticky="NSEW", columnspan=s, ipadx=4, ipady=4, padx=4, pady=4)

    def window(self) -> Window:
        return self._window

    def rd(self) -> T:
        return self._window.runtime_data()

    def active(self) -> bool:
        return self._window.active()

    def root(self) -> _CTk:
        return self._window.root()

    def show(self) -> None:
        self._window.show()

    def kill(self) -> None:
        self._window.kill()



from datetime import datetime
from time import time
from typing import Callable

from customtkinter import CTkButton, CTkLabel, IntVar, StringVar, CTkSegmentedButton
from keyboard import add_hotkey

from leads import *
from leads.comm import *
from leads_gui import *
from leads_vec.__version__ import __version__


class CustomRuntimeData(RuntimeData):
    m1_mode: int = 0
    m3_mode: int = 0
    control_system_switch_changed: bool = False


def make_system_switch(ctx: LEADS, system: SystemLiteral, runtime_data: RuntimeData) -> Callable[[], None]:
    def switch() -> None:
        ctx.plugin(system).enabled(not ctx.plugin(system).enabled())
        runtime_data.control_system_switch_changed = True

    return switch


def main() -> int:
    cfg = get_config(Config)
    ctx = LEADS[SRWDataContainer if cfg.srw_mode else DRWDataContainer](srw_mode=cfg.srw_mode)
    ctx.plugin(SystemLiteral.DTCS, DTCS())
    ctx.plugin(SystemLiteral.ABS, ABS())
    ctx.plugin(SystemLiteral.EBI, EBI())
    ctx.plugin(SystemLiteral.ATBS, ATBS())
    SFT.bind_context(ctx)
    window = Window(cfg.width,
                    cfg.height,
                    cfg.refresh_rate,
                    CustomRuntimeData(),
                    fullscreen=cfg.fullscreen,
                    no_title_bar=cfg.no_title_bar)
    m1 = StringVar(window.root(), "")
    m2 = IntVar(window.root(), 0)
    m3 = StringVar(window.root(), "")
    ecs = StringVar(window.root(), "STANDARD")

    def render(manager: ContextManager):
        def switch_m1_mode():
            manager.rd().m1_mode = (manager.rd().m1_mode + 1) % 2

        def switch_m3_mode():
            manager.rd().m3_mode = (manager.rd().m3_mode + 1) % 3

        manager["m1"] = CTkButton(manager.root(), textvariable=m1, command=switch_m1_mode,
                                  font=("Arial", cfg.font_size_small))
        manager["m2"] = CTkButton(manager.root(), textvariable=m2, state="disabled",
                                  font=("Arial", cfg.font_size_x_large))
        manager["m3"] = CTkButton(manager.root(), textvariable=m3, command=switch_m3_mode,
                                  font=("Arial", cfg.font_size_medium))

        manager["comm_status"] = CTkLabel(manager.root(), text="COMM OFFLINE", text_color="gray",
                                          font=("Arial", cfg.font_size_small))

        i = 0
        for system in SystemLiteral:
            i += 1
            system_lower = system.lower()
            manager[system_lower + "_status"] = CTkLabel(manager.root(), text=system + " READY", text_color="green",
                                                         font=("Arial", cfg.font_size_small))
            add_hotkey(str(i), switch := make_system_switch(ctx, SystemLiteral(system), manager.rd()))
            manager[system_lower] = CTkButton(manager.root(), text=system + " ON", command=switch,
                                              font=("Arial", cfg.font_size_small))

        manager["time_lap"] = CTkButton(manager.root(), text="Time Lap", command=ctx.time_lap,
                                        font=("Arial", cfg.font_size_small))

        def hazard():
            ctx.hazard(not ctx.hazard())
            manager["hazard"].configure(image=Hazard(color=Color.RED if ctx.hazard() else None))

        manager["hazard"] = CTkButton(manager.root(), text="", image=Hazard(), command=hazard)

        def switch_ecs_mode(mode):
            manager["ecs"].configure(selected_color=(c := "green" if (ecs_mode := ECSMode[mode]) < 2 else "red"),
                                     selected_hover_color=c)
            ctx.ecs_mode(ecs_mode)
            manager.rd().control_system_switch_changed = True

        manager["ecs"] = CTkSegmentedButton(manager.root(), values=["STANDARD", "AGGRESSIVE", "SPORT", "OFF"],
                                            variable=ecs, command=switch_ecs_mode, font=("Arial", cfg.font_size_small))

    uim = initialize(window, render, ctx, get_controller(MAIN_CONTROLLER))

    class CommCallback(Callback):
        def on_fail(self, service: Service, error: Exception) -> None:
            L.error("Comm server error: " + repr(error))

        def on_receive(self, service: Service, msg: bytes) -> None:
            if msg == b"time_lap":
                ctx.time_lap()

    uim.rd().comm = start_server(create_server(cfg.comm_port, CommCallback()), True)

    def format_lap_time(t: int) -> str:
        return f"{(t := int(t * .001)) // 60} MIN {t % 60} SEC"

    class CustomListener(EventListener):
        def on_push(self, e: DataPushedEvent) -> None:
            d = e.data.to_dict()
            d["speed_trend"] = ctx.get_speed_trend()
            d["lap_times"] = ctx.get_lap_time_list()
            uim.rd().comm_notify(d)

        def post_update(self, e: UpdateEvent) -> None:
            d = e.context.data()
            if uim.rd().m1_mode == 0:
                lap_time_list = ctx.get_lap_time_list()
                m1.set("LAP TIMES\n\n" + ("No Lap Timed" if len(lap_time_list) < 1 else "\n".join(map(format_lap_time,
                                                                                                      lap_time_list))))
            else:
                m1.set(f"VeC {__version__.upper()}\n\n"
                       f"{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}\n"
                       f"{(duration := int(time()) - uim.rd().start_time) // 60} MIN {duration % 60} SEC\n"
                       f"{(m := d.mileage):.1f} KM - {m * 3600 / duration:.1f} KM / H\n\n"
                       f"{"SRW MODE" if cfg.srw_mode else "DRW MODE"}\n"
                       f"{cfg.refresh_rate} FPS\n"
                       f"{ip[-1] if len(ip := my_ip_addresses()) > 0 else "NOT FOUND"}:{uim.rd().comm.port()}")
            m2.set(int(d.speed))
            if uim.rd().m3_mode == 0:
                m3.set(f"{d.voltage:.1f} V")
            elif uim.rd().m3_mode == 1:
                m3.set("G Force")
            else:
                m3.set(f"Speed Trend\n"
                       f"{(st := ctx.get_speed_trend() * 10):.1f} {"↑" if st > 0 else "↓"}")
            if uim.rd().comm.num_connections() < 1:
                uim["comm_status"].configure(text="COMM OFFLINE", text_color="gray")
            else:
                uim["comm_status"].configure(text="COMM ONLINE", text_color=["black", "white"])
            if uim.rd().control_system_switch_changed:
                for system in SystemLiteral:
                    system_lowercase = system.lower()
                    if ctx.plugin(SystemLiteral(system)).enabled():
                        uim[system_lowercase].configure(text=system + " ON")
                    else:
                        uim[system_lowercase].configure(text=system + " OFF")
                        uim[system_lowercase + "_status"].configure(text=system + " OFF", text_color=("black", "white"))
                uim.rd().control_system_switch_changed = False

        def on_intervene(self, e: InterventionEvent) -> None:
            if e.system in SystemLiteral:
                uim[e.system.lower() + "_status"].configure(text=e.system + " INTEV", text_color="red")

        def post_intervene(self, e: InterventionEvent) -> None:
            if e.system in SystemLiteral:
                uim[e.system.lower() + "_status"].configure(text=e.system + " READY", text_color="green")

        def on_suspend(self, e: SuspensionEvent) -> None:
            if e.system in SystemLiteral:
                uim[e.system.lower() + "_status"].configure(text=e.system + " SUSPD", text_color="gray")

        def post_suspend(self, e: SuspensionEvent) -> None:
            if e.system in SystemLiteral:
                uim[e.system.lower() + "_status"].configure(text=e.system + " READY", text_color="green")

    ctx.set_event_listener(CustomListener())
    uim["battery_fault"] = CTkLabel(uim.root(), text="")
    uim["ecs_fault"] = CTkLabel(uim.root(), text="")
    uim["gps_fault"] = CTkLabel(uim.root(), text="")
    uim["motor_fault"] = CTkLabel(uim.root(), text="")
    uim["wheel_speed_fault"] = CTkLabel(uim.root(), text="")

    def on_fail(_, e: SuspensionEvent) -> None:
        if e.system == "ECS":
            uim["ecs_fault"].configure(image=ECS(color=Color.RED))
        elif e.system == "BATT":
            uim["battery_fault"].configure(image=Battery(color=Color.RED))
        elif e.system == "MOTOR":
            uim["motor_fault"].configure(image=Motor(color=Color.RED))
        elif e.system == "WSC":
            uim["wheel_speed_fault"].configure(image=Speed(color=Color.RED))
        elif e.system == "GPS":
            uim["gps_fault"].configure(image=Satellite(color=Color.RED))

    SFT.on_fail = on_fail

    def on_recover(_, e: SuspensionEvent) -> None:
        if e.system == "ECS":
            uim["ecs_fault"].configure(image=None)
        elif e.system == "BATT":
            uim["battery_fault"].configure(image=None)
        elif e.system == "MOTOR":
            uim["motor_fault"].configure(image=None)
        elif e.system == "WSC":
            uim["wheel_speed_fault"].configure(image=None)
        elif e.system == "GPS":
            uim["gps_fault"].configure(image=None)

    SFT.on_recover = on_recover
    layout = [
        ["m1", "m2", "m3"],
        ["dtcs_status", "abs_status", "ebi_status", "atbs_status", "comm_status"],
        list(map(lambda s: s.lower(), SystemLiteral)),
        ["ecs"],
        ["time_lap", "hazard"],
        ["battery_fault", "ecs_fault", "gps_fault", "motor_fault", "wheel_speed_fault"]
    ]
    uim.layout(layout)
    placeholder_row = len(layout)
    CTkLabel(uim.root(), text="").grid(row=placeholder_row, column=0)
    uim.root().grid_rowconfigure(0, weight=1)
    uim.root().grid_rowconfigure(placeholder_row, weight=2)
    initialize_main()
    uim.show()
    uim.rd().comm_kill()
    return 0



from importlib.util import find_spec as _find_spec

if not _find_spec("PIL"):
    raise ImportError("Please install `Pillow` to run this module\n>>>pip install Pillow")
if not _find_spec("customtkinter"):
    raise ImportError("Please install `customtkinter` to run this module\n>>>pip install customtkinter")

from os.path import abspath as _abspath
from typing import Callable as _Callable, Any as _Any
from customtkinter import set_default_color_theme as _set_default_color_theme

from leads import LEADS as _LEADS, Controller as _Controller
from leads.comm import Server as _Server, create_server as _create_server
from leads_gui.prototype import *
from leads_gui.runtime import *
from leads_gui.config import *
from leads_gui.icons import *

_set_default_color_theme(_abspath(__file__)[:-11] + "assets/leads-theme.json")


def initialize(window: Window,
               render: _Callable[[ContextManager], None],
               leads: _LEADS[_Any],
               main_controller: _Controller) -> ContextManager:
    ctx = ContextManager(window)
    render(ctx)

    def on_refresh(_) -> None:
        leads.push(main_controller.read())
        leads.update()

    window.set_on_refresh(on_refresh)
    return ctx


def initialize_comm_server(window: Window,
                           render: _Callable[[ContextManager], None],
                           server: _Server = _create_server()) -> None:
    ctx = ContextManager(window)
    render(ctx)
    server.start(True)

    def on_close(_) -> None:
        server.kill()

    window.set_on_close(on_close)



from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from threading import Thread as _Thread
from typing import Any as _Any, override as _override


class Device(object):
    def __init__(self, *pins: int | str) -> None:
        self._tag: str = ""
        self._parent_tags: list[str] = []
        self._pins: tuple[int | str, ...] = pins

    def tag(self, tag: str | None = None) -> str | None:
        if tag:
            self._tag = tag
        else:
            return self._tag

    def parent_tags(self, parent_tags: list[str] | None = None) -> list[str] | None:
        if parent_tags:
            if len(self._parent_tags) > 0:
                raise RuntimeError("Duplicated initialization")
            self._parent_tags = parent_tags
        else:
            return self._parent_tags[:]

    def pins_check(self, required_num: int) -> None:
        if len(self._pins) != required_num:
            raise ValueError(f"`{self.__class__.__name__}` only takes in {required_num} pins")

    def initialize(self, *parent_tags: str) -> None:
        pass

    def read(self) -> _Any:
        raise NotImplementedError

    def write(self, payload: _Any) -> None:
        raise NotImplementedError

    def update(self, data: _Any) -> None:
        raise NotImplementedError

    def close(self) -> None:
        pass


class ShadowDevice(Device, metaclass=_ABCMeta):
    def __init__(self, *pins: int | str) -> None:
        super().__init__(*pins)
        self._shadow_thread: _Thread | None = None

    @_abstractmethod
    def loop(self) -> None:
        raise NotImplementedError

    def run(self) -> None:
        while True:
            self.loop()

    @_override
    def initialize(self, *parent_tags: str) -> None:
        self._shadow_thread = _Thread(name=str(id(self)) + " shadow", target=self.run)
        self._shadow_thread.start()



from typing import override as _override

from pynmea2 import parse as _parse
from pynmea2.types.talker import TalkerSentence as _TalkerSentence, GGA as _GGA, GSA as _GSA
from serial import Serial as _Serial

from leads import Device as _Device, SFT as _SFT
from leads.comm import Entity as _Entity, Callback as _Callback, Service as _Service
from leads_comm_serial import SerialConnection as _SerialConnection


class _GPSCallback(_Callback):
    def __init__(self, receiver: _Device) -> None:
        self.receiver: _Device = receiver

    @_override
    def on_receive(self, service: _Service, msg: bytes) -> None:
        assert isinstance(service, GPSReceiver)
        self.receiver.update(_parse(msg.decode()))

    @_override
    def on_fail(self, service: _Service, error: Exception) -> None:
        assert isinstance(service, GPSReceiver)
        _SFT.fail(service, error)


class GPSReceiver(_Device, _Entity):
    """
    Supports:
    - Any USB (serial) GPS receiver with NMEA 0183 output
    """
    def __init__(self, port: str, baud_rate: int = 9600) -> None:
        _Device.__init__(self, port)
        _Entity.__init__(self, -1, _GPSCallback(self))
        self._serial: _Serial = _Serial()
        self._serial.port = port
        self._serial.baudrate = baud_rate
        self._connection: _SerialConnection | None = None
        self._latitude: float = 0
        self._longitude: float = 0
        self._valid: bool = True

    @_override
    def port(self) -> str:
        return self._serial.port

    @_override
    def initialize(self, *parent_tags: str) -> None:
        self.start(True)
        super().initialize(*parent_tags)

    @_override
    def update(self, data: _TalkerSentence) -> None:
        if isinstance(data, _GGA):
            self._latitude = float(data.latitude)
            self._longitude = float(data.longitude)
        elif isinstance(data, _GSA):
            if (v := data.is_valid) and not self._valid:
                _SFT.recover(self)
            elif not v and self._valid:
                _SFT.fail(self, "No fix")
            self._valid = v

    @_override
    def read(self) -> [bool, float, float]:
        """
        :return: [validity, latitude, longitude]
        """
        return self._valid, self._latitude, self._longitude

    @_override
    def run(self) -> None:
        self.callback.on_initialize(self)
        self._serial.open()
        self.callback.on_connect(self, connection := _SerialConnection(self, self._serial, self._serial.port,
                                                                       separator=b"\n"))
        self._connection = connection
        self._stage(connection)

    @_override
    def kill(self) -> None:
        if self._connection:
            self._connection.close()

    @_override
    def close(self) -> None:
        self.kill()



from argparse import ArgumentParser as _ArgumentParser
from os import mkdir as _mkdir, chmod as _chmod
from os.path import exists as _exists
from sys import exit as _exit, version as _version

from leads import register_controller as _register_controller, MAIN_CONTROLLER as _MAIN_CONTROLLER, \
    L as _L, load_config as _load_config, register_config as _register_config, \
    get_config as _get_config
from leads_gui import get_system_platform as _get_system_platform, Config as _Config

if __name__ == "__main__":
    parser = _ArgumentParser(prog="LEADS VeC",
                             description="Lightweight Embedded Assisted Driving System VeC",
                             epilog="ProjectNeura: https://projectneura.org"
                                    "GitHub: https://github.com/ProjectNeura/LEADS")
    parser.add_argument("action", choices=("info", "run"))
    parser.add_argument("-r", "--register", choices=("systemd", "config"), default=None, help="service to register")
    parser.add_argument("-c", "--config", default=None, help="specified configuration file")
    args = parser.parse_args()
    if args.action == "info":
        from leads_vec.__version__ import __version__

        _L.info("LEADS Version: " + __version__,
                "System Platform: " + _get_system_platform(),
                "Python Version: " + _version,
                sep="\n")
        _exit()
    if args.register == "systemd":
        if _get_system_platform() != "linux":
            _exit("Error: Unsupported operating system")
        if not _exists("/usr/local/leads/config.json"):
            _L.info("Config file not found. Creating \"/usr/local/leads/config.json\"...")
            _mkdir("/usr/local/leads")
            with open("/usr/local/leads/config.json", "w") as f:
                f.write(str(_get_config(_Config)))
            _L.info("Using \"/usr/local/leads/config.json\"")
        _chmod("/usr/local/leads/config.json", 777)
        from ._bootloader import create_service

        create_service()
        _L.info("Service registered")
    elif args.register == "config":
        if _exists("config.json"):
            r = input("\"config.json\" already exists. Overwrite? (y/N) >>>").lower()
            if r.lower() != "y":
                _exit("Error: Aborted")
        with open("config.json", "w") as f:
            f.write(str(_get_config(_Config)))
        _L.info("Configuration file saved to \"config.json\"")
    _register_config(config := _load_config(args.config, _Config) if args.config else None)
    _L.debug("Configuration loaded:", str(config))
    from leads_vec.cli import main

    try:
        from leads_vec.controller import _
    except ImportError as e:
        _L.debug(repr(e))
        _L.warn("`leads_vec.controller` is not available, using emulation module instead...")
        try:
            from leads_emulation import SRWRandom as _Controller

            _register_controller(_MAIN_CONTROLLER, _Controller())
        except ImportError:
            raise ImportError("At least one adapter has to be installed")

    _exit(main())



import sys as _sys
from types import FrameType as _FrameType

if hasattr(_sys, "_getframe"):
    def currentframe() -> _FrameType:
        return getattr(_sys, "_getframe")(1)
else:
    def currentframe() -> _FrameType:
        try:
            raise Exception
        except Exception as exc:
            return exc.__traceback__.tb_frame.f_back



from typing import Any as _Any, Callable as _Callable, Sequence as _Sequence

from leads.dt.controller import Controller
from leads.dt.device import Device
from leads.dt.predefined_tags import MAIN_CONTROLLER

_controllers: dict[str, Controller] = {}
_devices: dict[str, Device] = {}


def controller(tag: str,
               parent: str | None = None,
               args: tuple[_Any, ...] = (),
               kwargs: dict[str, _Any] | None = None) -> _Callable[[type], None]:
    if not kwargs:
        kwargs = {}

    def _(target: type) -> None:
        if not issubclass(target, Controller):
            raise TypeError("Controllers must inherit from `Controller`")
        register_controller(tag, target(*args, **kwargs), parent)

    return _


def device(tag: str | _Sequence[str],
           parent: str | _Sequence[str],
           args: tuple[_Any, ...] | list[tuple[_Any, ...]] = (),
           kwargs: dict[str, _Any] | list[dict[str, _Any]] | None = None) -> _Callable[[type], None]:
    if isinstance(tag, str):
        tag = [tag]
    n = len(tag)
    if isinstance(parent, str):
        parent = [parent] * n
    if isinstance(args, tuple):
        args = [args] * n
    if not kwargs:
        kwargs = [{}] * n
    elif isinstance(kwargs, dict):
        kwargs = [kwargs] * n

    def _(target: type) -> None:
        if not issubclass(target, Device):
            raise TypeError("Devices must inherit from `Device`")
        for i in range(len(tag)):
            _register_device(target, tag[i], _controllers[parent[i]], args[i], kwargs[i])

    return _


def register_controller(tag: str, c: Controller, parent: str | None = None) -> None:
    if tag in _controllers:
        raise RuntimeError(f"Cannot register: tag \"{tag}\" is already used")
    if parent:
        c.parent_tags((p := _controllers[parent]).parent_tags() + c.parent_tags())
        p.device(tag, c)
    _controllers[tag] = c


def get_controller(tag: str) -> Controller:
    return _controllers[tag]


def _register_device(prototype: type,
                     tag: str,
                     parent: Controller,
                     args: tuple[_Any, ...],
                     kwargs: dict[str, _Any]) -> None:
    instance = prototype(*args, **kwargs)
    parent.parent_tags(parent.parent_tags())
    parent.device(tag, instance)
    _devices[tag] = instance


def get_device(tag: str) -> Device:
    return _devices[tag]


def initialize_main() -> None:
    get_controller(MAIN_CONTROLLER).initialize()



from leads import device, controller, MAIN_CONTROLLER, get_controller, WHEEL_SPEED_CONTROLLER, SRWDataContainer, \
    DRWDataContainer, LEFT_FRONT_WHEEL_SPEED_SENSOR, RIGHT_FRONT_WHEEL_SPEED_SENSOR, Controller, \
    CENTER_REAR_WHEEL_SPEED_SENSOR, LEFT_REAR_WHEEL_SPEED_SENSOR, RIGHT_REAR_WHEEL_SPEED_SENSOR, get_config, \
    mark_system, POWER_CONTROLLER, ODOMETER, GPS_RECEIVER
from leads_arduino import ArduinoMicro, WheelSpeedSensor, ArduinoCallback, VoltageSensor, ConcurrentOdometer
from leads_gui import Config
from leads_raspberry_pi import GPSReceiver

config = get_config(Config)
BAUD_RATE: int = config.get("baud_rate", 9600)
POWER_CONTROLLER_PORT: str = config.get("power_controller_port", "COM4")
WHEEL_SPEED_CONTROLLER_PORT: str = config.get("wheel_speed_controller_port", "COM3")
GPS_RECEIVER_PORT: str = config.get("gps_receiver_port", "COM5")
FRONT_WHEEL_DIAMETER: float = config.get("front_wheel_diameter", 20)  # 20 inches
REAR_WHEEL_DIAMETER: float = config.get("rear_wheel_diameter", 20)  # 20 inches
THROTTLE_PEDAL_PIN: int = config.get("throttle_pedal_pin", 2)
BRAKE_PEDAL_PIN: int = config.get("brake_pedal_pin", 3)
VOLTAGE_SENSOR_PIN: int = config.get("voltage_sensor_pin", 4)


@controller(MAIN_CONTROLLER)
class VeCController(Controller):
    def read(self) -> SRWDataContainer | DRWDataContainer:
        r = get_controller(WHEEL_SPEED_CONTROLLER).read()
        universal = {
            "voltage": get_controller(POWER_CONTROLLER).read(),
            "mileage": self.device(ODOMETER).read(),
            "gps_valid": (coords := self.device(GPS_RECEIVER).read())[0],
            "latitude": coords[1],
            "longitude": coords[2]
        }
        return SRWDataContainer(**r, **universal) if config.srw_mode else DRWDataContainer(**r, **universal)


@controller(POWER_CONTROLLER, MAIN_CONTROLLER, (POWER_CONTROLLER_PORT, ArduinoCallback(POWER_CONTROLLER), BAUD_RATE))
class PowerController(ArduinoMicro):
    def initialize(self, *parent_tags: str) -> None:
        mark_system(self, "POWER", "BATT", "MOTOR")
        super().initialize(*parent_tags)

    def read(self) -> float:
        return self.device("vot").read()


@device("vot", POWER_CONTROLLER)
class BatteryVoltageSensor(VoltageSensor):
    def initialize(self, *parent_tags: str) -> None:
        mark_system(self, "POWER", "BATT")


@controller(WHEEL_SPEED_CONTROLLER, MAIN_CONTROLLER, (
        WHEEL_SPEED_CONTROLLER_PORT, ArduinoCallback(WHEEL_SPEED_CONTROLLER), BAUD_RATE
))
class WheelSpeedController(ArduinoMicro):
    def initialize(self, *parent_tags: str) -> None:
        mark_system(self, "WSC", "ECS")
        super().initialize(*parent_tags)

    def read(self) -> dict[str, float]:
        lfws = self.device(LEFT_FRONT_WHEEL_SPEED_SENSOR).read()
        rfws = self.device(RIGHT_FRONT_WHEEL_SPEED_SENSOR).read()
        front_wheel_speed = (lfws + rfws) * .5
        return {
            "min_speed": min(lfws, rfws, rws := self.device(CENTER_REAR_WHEEL_SPEED_SENSOR).read()),
            "front_wheel_speed": front_wheel_speed,
            "rear_wheel_speed": rws
        } if config.srw_mode else {
            "min_speed": min(lfws, rfws, lrws := self.device(LEFT_REAR_WHEEL_SPEED_SENSOR).read(),
                             rrws := self.device(RIGHT_REAR_WHEEL_SPEED_SENSOR).read()),
            "front_wheel_speed": front_wheel_speed,
            "left_rear_wheel_speed": lrws,
            "right_rear_wheel_speed": rrws
        }


@device(ODOMETER, MAIN_CONTROLLER)
class AverageOdometer(ConcurrentOdometer):
    def read(self) -> float:
        return super().read() / (3 if config.srw_mode else 4)


@device(*((
        (LEFT_FRONT_WHEEL_SPEED_SENSOR,
         RIGHT_FRONT_WHEEL_SPEED_SENSOR,
         CENTER_REAR_WHEEL_SPEED_SENSOR),
        WHEEL_SPEED_CONTROLLER,
        [(FRONT_WHEEL_DIAMETER, ODOMETER),
         (FRONT_WHEEL_DIAMETER, ODOMETER),
         (REAR_WHEEL_DIAMETER, ODOMETER)
         ]) if config.srw_mode else (
        (LEFT_FRONT_WHEEL_SPEED_SENSOR,
         RIGHT_FRONT_WHEEL_SPEED_SENSOR,
         LEFT_REAR_WHEEL_SPEED_SENSOR,
         RIGHT_REAR_WHEEL_SPEED_SENSOR),
        WHEEL_SPEED_CONTROLLER,
        [(FRONT_WHEEL_DIAMETER, ODOMETER),
         (FRONT_WHEEL_DIAMETER, ODOMETER),
         (REAR_WHEEL_DIAMETER, ODOMETER),
         (REAR_WHEEL_DIAMETER, ODOMETER)]
)))
class WheelSpeedSensors(WheelSpeedSensor):
    def initialize(self, *parent_tags: str) -> None:
        mark_system(self, "WSC", "ECS")
        super().initialize(*parent_tags)


# @device((THROTTLE_PEDAL, BRAKE_PEDAL), MAIN_CONTROLLER, [(THROTTLE_PEDAL_PIN, False), (BRAKE_PEDAL_PIN, True)])
# class Pedals(Pedal):
#     def __init__(self, pin: int, brake: bool) -> None:
#         super().__init__(pin)
#         self._brake: bool = brake


@device(GPS_RECEIVER, MAIN_CONTROLLER, (GPS_RECEIVER_PORT,))
class GPS(GPSReceiver):
    def initialize(self, *parent_tags: str) -> None:
        mark_system(self, "GPS")
        super().initialize(*parent_tags)


_ = None  # null export



from json import loads
from os import mkdir
from os.path import abspath, exists
from typing import Any

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from leads import *
from leads.comm import *
from leads.data_persistence import *
from leads_gui import *

config = get_config(Config)
if not exists(config.data_dir):
    mkdir(config.data_dir)
    L.info(f"Data dir \"{abspath(config.data_dir)}\" created")

data_record: DataPersistence[DataContainer] = DataPersistence(max_size=1, compressor=lambda o, s: o[-s:])
time_stamp_record: DataPersistence[int] = DataPersistence(config.data_dir + "/time_stamp.csv",
                                                          persistence=config.enable_data_persistence,
                                                          max_size=2000)
speed_record: DataPersistence[float] = DataPersistence(config.data_dir + "/speed.csv",
                                                       persistence=config.enable_data_persistence,
                                                       max_size=2000)
gps_record: DataPersistence[Vector[float]] = DataPersistence(config.data_dir + "/gps.csv",
                                                             persistence=config.enable_data_persistence,
                                                             max_size=2000)


class CommCallback(Callback):
    def on_connect(self, service: Service, connection: Connection) -> None:
        L.debug("Connected")

    def on_fail(self, service: Service, error: Exception) -> None:
        L.error("Comm client error: " + repr(error))

    def on_receive(self, service: Service, msg: bytes) -> None:
        d = loads(msg.decode())
        data_record.append(d)
        time_stamp_record.append(d["t"])
        speed_record.append(d["front_wheel_speed"])
        gps_record.append(Vector(d["latitude"], d["longitude"]))

    def on_disconnect(self, service: Service, connection: Connection) -> None:
        time_stamp_record.close()
        speed_record.close()


client = start_client(config.comm_addr, create_client(config.comm_port, CommCallback()), True)

app = FastAPI(title="LEADS VeC Remote Controller")

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["GET"],
    allow_headers=["*"],
)


@app.get("/")
async def index() -> str:
    return "LEADS VeC Remote Controller"


@app.get("/current")
async def current() -> dict[str, Any]:
    return data_record[-1]


@app.get("/time_stamp")
async def time_stamp() -> list[int]:
    return time_stamp_record.to_list()


@app.get("/speed")
async def speed() -> list[float]:
    return speed_record.to_list()


@app.get("/time_lap")
async def time_lap() -> str:
    client.send(b"time_lap")
    return "done"



from datetime import datetime as _datetime
from enum import IntEnum as _IntEnum
from threading import Lock as _Lock

from leads.os import currentframe


class Level(_IntEnum):
    DEBUG: int = 0x00
    INFO: int = 0x01
    WARN: int = 0x02
    ERROR: int = 0x03


class Logger(object):
    REGULAR: int = 0
    BOLD: int = 1
    ITALIC: int = 3
    UNDERLINED: int = 4
    INVERSE: int = 7

    BLACK: int = 30
    RED: int = 31
    GREEN: int = 32
    YELLOW: int = 33
    BLUE: int = 34
    PURPLE: int = 35
    CYAN: int = 36
    WHITE: int = 37

    def __init__(self) -> None:
        self._debug_level: Level = Level.DEBUG
        self._lock: _Lock = _Lock()

    def debug_level(self, debug_level: Level | None = None) -> Level | None:
        if debug_level is None:
            return self._debug_level
        self._debug_level = debug_level

    @staticmethod
    def mark(msg: str, level: Level) -> str:
        return f"[{repr(level)[1:-1]}] [{currentframe().f_back.f_back.f_code.co_name}] [{_datetime.now()}] {msg}"

    @staticmethod
    def format(msg: str, font: int, color: int | None, background: int | None) -> str:
        return f"\033[{font}{f";{color}" if color else ""}{f";{background + 10}" if background else ""}m{msg}\033[0m"

    def print(self, msg: str, level: int) -> None:
        self._lock.acquire()
        try:
            if self._debug_level <= level:
                print(msg)
        finally:
            self._lock.release()

    def info(self, *msg: str, sep: str = " ", end: str = "\n",
             f: tuple[int, int | None, int | None] = (REGULAR, None, None)) -> None:
        self.print(self.format(self.mark(sep.join(msg) + end, level=Level.INFO), *f), Level.INFO)

    def debug(self, *msg: str, sep: str = " ", end: str = "\n",
              f: tuple[int, int | None, int | None] = (REGULAR, YELLOW, None)) -> None:
        self.print(self.format(self.mark(sep.join(msg) + end, level=Level.DEBUG), *f), Level.DEBUG)

    def warn(self, *msg: str, sep: str = " ", end: str = "\n",
             f: tuple[int, int | None, int | None] = (REGULAR, RED, None)) -> None:
        self.print(self.format(self.mark(sep.join(msg) + end, level=Level.WARN), *f), Level.WARN)

    def error(self, *msg: str, sep: str = " ", end: str = "\n",
              f: tuple[int, int | None, int | None] = (REGULAR, RED, None)) -> None:
        self.print(self.format(self.mark(sep.join(msg) + end, Level.ERROR), *f), Level.ERROR)


L: Logger = Logger()



from argparse import ArgumentParser as _ArgumentParser

from uvicorn import run

from leads import register_config as _register_config, load_config as _load_config
from leads_gui import Config as _Config

if __name__ == "__main__":
    parser = _ArgumentParser(prog="LEADS VeC RC",
                             description="Lightweight Embedded Assisted Driving System VeC Remote Controller",
                             epilog="GitHub: https://github.com/ProjectNeura/LEADS")
    parser.add_argument("-c", "--config", default=None, help="specified configuration file")
    parser.add_argument("-p", "--port", type=lambda p: int(p), default=8000, help="server port")
    args = parser.parse_args()
    _register_config(_load_config(args.config, _Config) if args.config else None)
    from leads_vec_rc.cli import app

    run(app, host="0.0.0.0", port=args.port, log_level="warning")



from enum import StrEnum as _StrEnum
from typing import Callable as _Callable

from PIL import Image as _Image
from customtkinter import CTkImage as _CTkImage

from leads import set_on_register_config as _set_on_register_config, OnRegisterConfig as _OnRegisterConfig, \
    get_config as _get_config
from leads_gui.config import Config
from leads_gui.system import _ASSETS_PATH

_ICONS_PATH: str = _ASSETS_PATH + "/icons"

_config: Config = _get_config(Config)


class Color(_StrEnum):
    BLACK: str = "black"
    WHITE: str = "white"
    RED: str = "red"


class _Icon(_Callable[[int, Color | None], _CTkImage]):
    def __init__(self, name: str) -> None:
        self._name: str = name

    def load_source(self, color: Color) -> _Image:
        return _Image.open(f"{_ICONS_PATH}/{self._name}-{color}.png")

    def __call__(self, size: int = _config.font_size_medium, color: Color | None = None) -> _CTkImage:
        return _CTkImage(self.load_source(color if color else Color.BLACK),
                         None if color else self.load_source(Color.WHITE),
                         size=(size, size))


Battery: _Icon = _Icon("battery")
Brake: _Icon = _Icon("brake")
ECS: _Icon = _Icon("ecs")
Engine: _Icon = _Icon("engine")
Hazard: _Icon = _Icon("hazard")
HighBeam: _Icon = _Icon("high-beam")
Light: _Icon = _Icon("light")
Motor: _Icon = _Icon("motor")
Satellite: _Icon = _Icon("satellite")
Speed: _Icon = _Icon("speed")


def _on_register_config(chain: _OnRegisterConfig[Config] | None) -> _OnRegisterConfig[Config]:
    def _(config: Config) -> None:
        global _config
        if chain:
            chain(config)
        _config = config

    return _


_set_on_register_config(_on_register_config)
if _c := _get_config(Config):
    _on_register_config(None)(_c)



from leads.comm.prototype import Callback
from leads.comm.server.server import Server


def create_server(port: int = 16900, callback: Callback = Callback()) -> Server:
    return Server(port, callback)


def start_server(target: Server = create_server(), parallel: bool = False) -> Server:
    return target.start(parallel)



from json import load as _load
from typing import TypeVar as _TypeVar, TextIO as _TextIO, Callable as _Callable, Any as _Any

from leads.config.template import ConfigTemplate
from leads.logger import Level, L
from leads.types import OnRegisterConfig, OnRegisterConfigChain

T = _TypeVar("T", bound=ConfigTemplate)


_config_instance: T | None = None

_on_register_config: OnRegisterConfig[T] = lambda c: L.debug_level(Level[c.debug_level])


def set_on_register_config(callback: OnRegisterConfigChain[T]) -> None:
    global _on_register_config
    _on_register_config = callback(_on_register_config)


def load_config(file: str | _TextIO, constructor: _Callable[[dict[str, _Any]], T]) -> T:
    if isinstance(file, str):
        with open(file) as f:
            return constructor(_load(f))
    return constructor(_load(file))


def register_config(config: T | None) -> None:
    global _config_instance
    if config:
        if _config_instance:
            raise RuntimeError("Another config is already registered")
        _on_register_config(config)
    _config_instance = config


def get_config(constructor: _Callable[[dict[str, _Any]], T]) -> T:
    if not _config_instance:
        return constructor({})
    return _config_instance



from typing import Callable as _Callable

from leads.context import ContextAssociated
from leads.dt import Device
from leads.event import SuspensionEvent
from leads.logger import L


def mark_system(device: Device, system: str, *related: str) -> None:
    if hasattr(device, "__device_system__"):
        setattr(device, "__device_system__", getattr(device, "__device_system__") + related)
    setattr(device, "__device_system__", [system, *related])


def read_marked_system(device: Device) -> list[str] | None:
    return getattr(device, "__device_system__") if hasattr(device, "__device_system__") else None


class SystemFailureTracer(ContextAssociated):
    def __init__(self) -> None:
        super().__init__()
        self.on_fail: _Callable[[Device, SuspensionEvent], None] = lambda _, __: None
        self.on_recover: _Callable[[Device, SuspensionEvent], None] = lambda _, __: None

    def fail(self, device: Device, error: str | Exception) -> None:
        if isinstance(error, Exception):
            error = repr(error)
        if not (systems := read_marked_system(device)):
            raise RuntimeWarning("No system marked for device " + str(device))
        for system in systems:
            self.on_fail(device, e := SuspensionEvent(context := self.require_context(), system, error))
            context.suspend(e)
            L.error(f"{system} error: {error}")

    def recover(self, device: Device) -> None:
        if not (systems := read_marked_system(device)):
            raise RuntimeWarning("System not marked for device " + str(device))
        for system in systems:
            self.on_recover(device, SuspensionEvent(self._context, system, "Recovered"))
            L.info(system + " recovered")


SFT: SystemFailureTracer = SystemFailureTracer()



from typing import override as _override

from RPi.GPIO import setup, OUT

from leads import Device as _Device


class DCMotor(_Device):
    @_override
    def initialize(self, *parent_tags: str) -> None:
        setup(self._pins[0], OUT)



from typing import override as _override

from leads.dt.device import Device


class Controller(Device):
    def __init__(self) -> None:
        super().__init__()
        self._devices: dict[str, Device] = {}

    def level(self) -> int:
        return len(self._parent_tags)

    def _attach_device(self, tag: str, device: Device) -> None:
        self._devices[tag] = device
        device.tag(tag)

    def devices(self) -> list[Device]:
        return list(self._devices.values())

    def device(self, tag: str, device: Device | None = None) -> Device | None:
        if device:
            self._attach_device(tag, device)
        else:
            return self._devices[tag]

    @_override
    def initialize(self, *parent_tags: str) -> None:
        for device in self._devices.values():
            device.initialize(*tuple(self._parent_tags), self._tag)



from typing import Callable as _Callable, SupportsInt as _SupportsInt, SupportsFloat as _SupportsFloat

type Number = int | float | _SupportsInt | _SupportsFloat
type OptionalSpeed = float | None
type Compressor[T] = _Callable[[list[T], int], list[T]]
type Stringifier[T] = _Callable[[T], str]
type RequiredData = list[str] | tuple[list[str], list[str]]
type OnRegisterConfig[T] = _Callable[[T], None]
type OnRegisterConfigChain[T] = _Callable[[OnRegisterConfig[T] | None], OnRegisterConfig[T]]



from typing import TypeVar as _TypeVar, Any as _Any, override as _override

from leads.context import Context
from leads.data import DataContainer
from leads.event import EventListener, DataPushedEvent, UpdateEvent, SuspensionEvent, InterventionEvent, \
    InterventionExitEvent
from leads.plugin import Plugin

T = _TypeVar("T", bound=DataContainer)


class LEADS(Context[T]):
    def __init__(self, event_listener: EventListener = EventListener(), *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._plugins: dict[str, Plugin] = {}
        self._event_listener: EventListener = event_listener

    def plugin(self, key: str, plugin: Plugin | None = None) -> Plugin | None:
        if plugin:
            plugin.bind_context(self)
            self._plugins[key] = plugin
            plugin.load()
        else:
            return self._plugins[key]

    def set_event_listener(self, event_listener: EventListener) -> None:
        self._event_listener = event_listener

    @_override
    def suspend(self, event: SuspensionEvent) -> None:
        self._event_listener.on_suspend(event)

    def _acquire_data(self, name: str, *systems: str, mandatory: bool = True) -> _Any | None:
        try:
            return getattr(self.data(), name)
        except AttributeError:
            if mandatory:
                for system in systems:
                    self.suspend(SuspensionEvent(self, system, f"no data for `{name}`"))

    @_override
    def push(self, data: T) -> None:
        self._event_listener.on_push(DataPushedEvent(self, data))
        super().push(data)
        self._event_listener.post_push(DataPushedEvent(self, data))

    @_override
    def intervene(self, event: InterventionEvent) -> None:
        if isinstance(event, InterventionExitEvent):
            self._event_listener.post_intervene(event)
        else:
            self._event_listener.on_intervene(event)

    @_override
    def update(self) -> None:
        self._event_listener.on_update(UpdateEvent(self))

        for key, plugin in self._plugins.items():
            if plugin.enabled():
                plugin.update({d: self._acquire_data(d, key) for d in plugin.required_data()})

        self._event_listener.post_update(UpdateEvent(self))

    @_override
    def time_lap(self) -> None:
        self.intervene(InterventionEvent(self, "LAP RECORDING"))
        super().time_lap()
        return self.intervene(InterventionExitEvent(self, "LAP RECORDING"))

    @_override
    def overwrite_throttle(self, force: float) -> float:
        self.intervene(InterventionEvent(self, "THROTTLE", force))
        try:
            return super().overwrite_throttle(force)
        finally:
            self.intervene(InterventionExitEvent(self, "THROTTLE", force))

    @_override
    def overwrite_brake(self, force: float) -> float:
        self.intervene(InterventionEvent(self, "BRAKE", force))
        try:
            return super().overwrite_brake(force)
        finally:
            self.intervene(InterventionExitEvent(self, "BRAKE", force))



from typing import Any as _Any

from leads.constant import ECSMode
from leads.context import Context, ContextAssociated
from leads.types import RequiredData


class Plugin(ContextAssociated):
    def __init__(self, required_data: RequiredData) -> None:
        """
        :param required_data: required data or (required data in srw mode, required data in drw mode)
        """
        super().__init__()
        self._required_data: RequiredData = required_data
        self.state: dict[str, _Any] = {}
        self._enabled: bool = True

    def __getitem__(self, key: str) -> _Any:
        return self.state[key]

    def __setitem__(self, key: str, value: _Any) -> None:
        self.state[key] = value

    def enabled(self, enabled: bool | None = None) -> bool | None:
        if enabled is None:
            return self.require_context().ecs_mode() != ECSMode.OFF and self._enabled
        self._enabled = enabled

    def required_data(self) -> list[str]:
        return (self._required_data if isinstance(self._required_data, list) else
                self._required_data[0] if self.require_context().srw_mode() else self._required_data[1])

    def load(self) -> None:
        self.on_load(self.require_context())

    def on_load(self, context: Context) -> None:
        pass

    def update(self, kwargs: dict[str, _Any]) -> None:
        self.on_update(self.require_context(), kwargs)

    def on_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        pass



from importlib.util import find_spec as _find_spec

if not _find_spec("RPi"):
    raise ImportError("Please install `RPi.GPIO` to run this module\n>>>apt install python3-rpi.gpio")
if not _find_spec("pynmea2"):
    raise ImportError("Please install `pynmea2` to run this module\n>>>pip install pynmea2")
if not _find_spec("serial"):
    raise ImportError("Please install `pyserial` to run this module\n>>>pip install pyserial")

from leads_raspberry_pi.gps_receiver import *
from leads_raspberry_pi.dc_motor import *



from typing import override as _override

from serial import Serial as _Serial

from leads import Controller as _Controller, get_controller as _get_controller, SFT as _SFT
from leads.comm import Entity as _Entity, Callback as _Callback, Service as _Service
from leads_comm_serial import SerialConnection as _SerialConnection


class ArduinoProto(_Controller, _Entity):
    """
    Supports:
    - Any arduino connected through a USB (serial) port
    """
    def __init__(self, port: str, callback: _Callback, baud_rate: int = 9600) -> None:
        _Controller.__init__(self)
        _Entity.__init__(self, -1, callback)
        self._serial: _Serial = _Serial()
        self._serial.port = port
        self._serial.baudrate = baud_rate
        self._connection: _SerialConnection | None = None

    @_override
    def port(self) -> str:
        return self._serial.port

    @_override
    def initialize(self, *parent_tags: str) -> None:
        self.start(True)
        super().initialize(*parent_tags)

    @_override
    def update(self, data: str) -> None:
        for d in self.devices():
            d.update(data)

    @_override
    def run(self) -> None:
        self.callback.on_initialize(self)
        self._serial.open()
        self.callback.on_connect(self, connection := _SerialConnection(self, self._serial, self._serial.port))
        self._connection = connection
        self._stage(connection)

    @_override
    def write(self, payload: bytes) -> None:
        if not self._connection:
            raise IOError("Target must be connected to perform this operation")
        self._connection.send(payload)

    @_override
    def kill(self) -> None:
        if self._connection:
            self._connection.close()

    @_override
    def close(self) -> None:
        self.kill()


class ArduinoCallback(_Callback):
    def __init__(self, tag: str) -> None:
        self._tag: str = tag

    @_override
    def on_receive(self, service: _Service, msg: bytes) -> None:
        _get_controller(self._tag).update(msg.decode())

    @_override
    def on_fail(self, service: _Service, error: Exception) -> None:
        assert isinstance(service, ArduinoProto)
        _SFT.fail(service, error)



from math import sin as _sin, pi as _pi
from random import randint as _randint
from typing import override as _override

from leads import Controller as _Controller, SRWDataContainer as _SRWDataContainer, \
    DRWDataContainer as _DRWDataContainer


class _EmulatedController(_Controller):
    def __init__(self,
                 minimum: int = 30,
                 maximum: int = 40,
                 skid_possibility: float = .1) -> None:
        super().__init__()
        self.minimum: int = minimum
        self.maximum: int = maximum
        self.skid_possibility: float = skid_possibility

    def generate_rear_wheel_speed(self, front_wheel_speed: float) -> float:
        return front_wheel_speed + int(_randint(1, int(1 / self.skid_possibility)) * self.skid_possibility)


class SRWRandom(_EmulatedController):
    @_override
    def read(self) -> _SRWDataContainer:
        return _SRWDataContainer(voltage=48.0,
                                 min_speed=(fws := _randint(self.minimum, self.maximum)),
                                 front_wheel_speed=fws,
                                 rear_wheel_speed=self.generate_rear_wheel_speed(fws))


class DRWRandom(_EmulatedController):
    @_override
    def read(self) -> _DRWDataContainer:
        return _DRWDataContainer(voltage=48.0,
                                 min_speed=(fws := _randint(self.minimum, self.maximum)),
                                 front_wheel_speed=fws,
                                 left_rear_wheel_speed=(rws := self.generate_rear_wheel_speed(fws)),
                                 right_rear_wheel_speed=rws)


class _SinController(_EmulatedController):
    def __init__(self,
                 minimum: int = 30,
                 maximum: int = 40,
                 skid_possibility: float = .1,
                 acceleration: float = .05) -> None:
        super().__init__(minimum, maximum, skid_possibility)
        self.acceleration: float = acceleration
        self.magnitude: int = int((maximum - minimum) * .5)
        self.offset: int = minimum
        self.counter: float = 0


class SRWSin(_SinController):
    @_override
    def read(self) -> _SRWDataContainer:
        try:
            return _SRWDataContainer(voltage=48.0,
                                     min_speed=(fws := (_sin(self.counter) + .5) * self.magnitude + self.offset),
                                     front_wheel_speed=fws,
                                     rear_wheel_speed=self.generate_rear_wheel_speed(fws))
        finally:
            self.counter = (self.counter + self.acceleration) % _pi


class DRWSin(_SinController):
    @_override
    def read(self) -> _DRWDataContainer:
        try:
            return _DRWDataContainer(voltage=48.0,
                                     min_speed=(fws := (_sin(self.counter) + .5) * self.magnitude + self.offset),
                                     front_wheel_speed=fws,
                                     left_rear_wheel_speed=(rws := self.generate_rear_wheel_speed(fws)),
                                     right_rear_wheel_speed=rws)
        finally:
            self.counter = (self.counter + self.acceleration) % _pi
